"""

# Managing different data sources

A `tessa.sources.sources.Source` ties together all the functionality and state related
to a source such as Yahoo or Coingecko.

`tessa.sources.rate_limiter.RateLimiter` takes care of rate limiting for a source.

All known sources are specified in `tessa.sources.sources_directory`.

"""

from .sourcetype import SourceType
from .sources import Source, get_source, get_all_sources, reset_rate_limiters



"""

# Working with Symbols

The main class is `tessa.symbol.symbol.Symbol`, which encapsulates all the information
and functionality around symbols. A `tessa.symbol.symbolcollection.SymbolCollection`
manages a collection of symbols including save and load functionality.

Example use:

```python
from tessa import Symbol, SymbolCollection
s1 = Symbol("MSFT")                 # will use "yahoo" as the default source
s1.price_latest()                   # get latest price
sc = SymbolCollection([s1])         # create collection
sc.add(Symbol("ethereum", source="coingecko"))       # add another symbol
sc.find_one("ethereum").price_graph(monthsback=600)  # graph of 10 past yearss
sc.save_yaml("my_symbols.yaml")     # save
```

`tessa.symbol.extended_symbol.ExtendedSymbol` shows how the `Symbol` class can be
extended.

"""

from .symbol import Symbol
from .extended_symbol import ExtendedSymbol
from .symbolcollection import SymbolCollection



"""
.. include:: ../README.md
"""

__version__ = "0.6.0"

from .sources import SourceType
from .price import (
    price_history,
    price_point,
    price_point_strict,
    price_latest,
)
from .search import search
from .symbol import Symbol, ExtendedSymbol, SymbolCollection



"""An example of how to extend the Symbol class with more information and functionality.
"""

from typing import Optional, List, Union
from dataclasses import dataclass, field
from . import Symbol
from .geo import CountryName, j2r


# Customizations to jurisdiction2region (remove or adapt according to your needs):
j2r.jurisdiction2region.update(
    {
        "CH": "CH",
        "CN": "CN",
        "EU": "EU",
        "irrelevant": "OT",
        "several": "OT",
        "unknown": "OT",
    }
)
j2r.region2name.update(
    {
        "CH": "Switzerland",
        "CN": "China",
        "AS": "Asia sans China",
    }
)


@dataclass
class ExtendedSymbol(Symbol):
    """ExtendedSymbol class. Adds more information and functionality to the Symbol
    class.
    """

    description: Optional[str] = None
    """Informational only."""
    # FIXME A design alternative could be to move the `description` attribute to
    # `Symbol` itself, rename it to `info` so it can be used for arbitrary additional
    # informaton around a symbol such as a description or similar.

    country: Optional[CountryName] = "united states"
    """Country of HQ / listing."""

    watch: bool = False
    """Whether the ticker can produce alerts."""

    delisted: bool = False
    """Used to filter certain symbols from analysis."""

    jurisdiction: str = "US"
    """main jurisdiction of the underlying asset(s) (not of the title representing the
    asset); default US, other examples: CN, EU, several, irrelevant, unknown
    """

    isin: Optional[str] = None
    """Informational only."""

    strategy: Union[str, List[str]] = field(default_factory=list)
    """One strategy or a list of several strategies. Some strategies I use:
      - F&F: Fire & Forget (implies HoldForGrowth)
      - HoldForGrowth
      - HoldForStability: low growth expected, but also low risk
      - HoldForDiversification: titles I find important for diversification
      - EnterIf: consider starting a position if prices for this title fall
      - SellIf: consider selling if prices high and/or liquidity needed
      - Sold: earlier holding that was sold at some point
      - Quarterly: quarterly invest
    """

    strategy_comments: Optional[str] = None
    """Additional comments re strategy."""

    region: str = field(init=False)
    """Will be set automatically."""

    def __post_init__(self) -> None:
        super().__post_init__()
        self.region = j2r.map_jurisdiction_to_region(self.jurisdiction)

    def get_strategy_string(self) -> str:
        """Return a nice string with the strategy including comments."""
        if isinstance(self.strategy, list):
            res = ", ".join(self.strategy)
        else:
            res = self.strategy
        if getattr(self, "strategy_comments", False):
            res += f" · {self.strategy_comments}"
        return res

    def price_graph(self, monthsback: int = 6) -> tuple:
        """Add some extra information to the price graph."""
        (from_date, fig, ax) = super().price_graph(monthsback)
        print(f"{self.name}")
        latestprice = self.price_latest().price
        print(f"Latest price: {latestprice:.2f}")
        hist = self.price_history()[0]
        maxprice = hist[hist.index > from_date].max().close
        print(f"Drop since max: {(latestprice - maxprice) / maxprice:2.0%}")
        return from_date, fig, ax



"""Everything related to search results, especially the `SearchResult` class."""

from __future__ import annotations
import functools
from typing import List, NamedTuple, Callable
import itertools
import re
from ..symbol import Symbol

# ----- The predicates used to sort and bucketize results based on a query -----

# Each matches_* function takes a `query` and a `Symbol` and returns an int denoting the
# match, where lower numbers signify better matches but 0 (=falsy) signifies not match.


def matches_entire_name_or_alias(query: str, symbol: Symbol) -> int:
    """Try to match the complete name or one of the aliases completely. E.g.:
    `matches_entire_name_or_alias("eth", Symbol("ETH")) > 0`.
    """
    query = query.lower()
    if query == symbol.name.lower():
        return 1
    if any(query == x.lower() for x in symbol.aliases if x):
        return 2
    return 0


def matches_word_boundary(query: str, symbol: Symbol) -> int:
    """Try to match a complete word in name, aliases, or query. E.g.:
    `matches_word_boundary("eth", Symbol("ETHW", aliases=["PoW ETH"])) > 0`.
    """
    query = query.lower()
    pattern = re.compile(rf"(^|[^a-z0-9]){query}([^a-z0-9]|$)", re.IGNORECASE)
    if re.search(pattern, symbol.name):
        return 1
    if re.search(pattern, "|".join(symbol.aliases)):
        return 2
    if re.search(pattern, str(symbol.query)):
        return 3
    return 0


def matches_inanyway(query: str, symbol: Symbol) -> int:
    """Try to match anywhere in name, aliases, or query. E.g.:
    `matches_inanyway("eth", Symbol("TOGETHER")) > 0`.
    """
    query = query.lower()
    if query in symbol.name.lower():
        return 1
    if query in ("|".join(symbol.aliases)).lower():
        return 2
    if query in str(symbol.query).lower():
        return 3
    return 0


# ----- Everything related to bucketizing symbols -----

Bucket = NamedTuple("Bucket", [("name", str), ("symbols", List[Symbol])])
"""A bucket is a set of results of a certain quality level in a larger search result. A
bucket has a name and a list of symbols that belong into that bucket.
"""


BUCKET_SPEC = [
    ("Matches entire name or alias ⭐", matches_entire_name_or_alias),
    ("Matches word boundary", matches_word_boundary),
    ("Matches in any way", matches_inanyway),
    # A bucket with the non-matches will be added by `bucketize`.
]


def bucketize(query: str, symbols: List[Symbol]) -> List[Bucket]:
    """Split a list of `Symbol`s into buckets based on how well a symbol matches the
    `query`.

    (We're not using `itertools.groupby` here bc we need empty lists for empty buckets,
    which `groupby` doesn't provide.)
    """
    buckets = []
    rest = symbols
    for name, matches_predicate in BUCKET_SPEC:
        matches_query_predicate = functools.partial(matches_predicate, query)
        hits = list(filter(matches_query_predicate, rest))
        rest = list(itertools.filterfalse(matches_query_predicate, rest))
        buckets.append(Bucket(name, hits))
    buckets.append(Bucket("Matches not (but was somehow still returned) 🤔", rest))
    return buckets


# ----- Sorting and removing duplicates -----


def create_sort_key_for_query(query: str) -> Callable:
    """Create a sort key function that classifies a `Symbol` based on `query`."""

    def symbol_sort_key(symbol: Symbol) -> int:
        """Cycle through the predicates in `BUCKET_SPEC` and use an offset to return an
        int based on the matching quality.
        """
        offset = 10
        for _, pred in BUCKET_SPEC:
            score = pred(query, symbol)
            if score > 0:
                return offset + score
            offset += 10
        return 999

    return symbol_sort_key


def remove_duplicates(symbols: List[Symbol]) -> List[Symbol]:
    """Return a list in which every symbol is unique. Two symbols are considered
    equivalent if they match in `query` and `source`. Regardless of whether their names
    are the same or different.

    Note further that several symbols can have the same name but still be considered
    different if they differ in any of `query` or `source`.
    """
    equality_key = lambda symbol: str((symbol.source, symbol.query))
    return [k for k, _ in itertools.groupby(sorted(symbols, key=equality_key))]


# ----- The SearchResult class -----


class SearchResult:
    """Manages a search result consisting of a list of `Symbol`s. Removes duplicates,
    bucketizes, sorts and offers methods to filter and print statistics.

    (Design note 1: `SearchResult` has no relationship with `SymbolCollection` because
    the purposes of the two classes are very different and because they have different
    equality definitions for symbols.

    Design note 2: `SearchResult` and `SymbolCollection` have different definitions of
    equality between symbols: `SearchResult` considers 2 symbols to be equal if they are
    equal in `query` and `source`, regardless of whether they have the same name or not.
    A `SymbolCollection` does not have a definition of equality, but it does enforce
    that names are unique.)

    Example use:

    ```python
    from tessa import search
    r = search("harmony")

    # Review results:
    r.p()

    # Get the 1 symbol from source "coingecko" in the best bucket (i.e., bucket 0):
    s = r.filter(source="coingecko").buckets[0].symbols[0]
    s.price_latest()
    ```
    """

    query: str
    """The query that produced the results here."""

    symbols: List[Symbol]
    """List of symbols in this result. Always deduplicated and sorted."""

    buckets: List[Bucket]
    """List of buckets after bucketizing the symbols."""

    filter_history: List[str]
    """The filters that have been applied to get to this result."""

    def __init__(self, query: str, symbols: List[Symbol]):
        self.query = query
        self.buckets = []
        self.filter_history = []
        self.symbols = []
        self.add_symbols(symbols)

    def add_symbols(self, symbols: List[Symbol]) -> SearchResult:
        """Add symbols. Removes duplicates, sorts, and bucketizes after adding."""
        self.symbols.extend(symbols)
        self.symbols = remove_duplicates(self.symbols)
        self.symbols.sort(key=create_sort_key_for_query(self.query))
        self.buckets = bucketize(self.query, self.symbols)
        return self

    def filter(self, **kwargs) -> SearchResult:
        """Filter for arbitrary attribute-value-pairs (e.g., `source="coingecko"` or
        `exchange="ebs"`) and return a new `SearchResult` with the results after the
        filtering. Also updates the filter history.
        """
        symbols = []
        for s in self.symbols:
            if all(getattr(s, k, "DOESNOTEXISTEVER") == v for k, v in kwargs.items()):
                symbols.append(s)
        newres = SearchResult(self.query, symbols)
        newres.filter_history = self.filter_history + [
            f"{k}={v}" for k, v in kwargs.items()
        ]
        return newres

    def __str__(self) -> str:
        """Return summary."""
        out = ""

        def s(line: str = "", indent=False) -> None:
            nonlocal out
            indent_str = "  " if indent else ""
            out += indent_str + line + "\n"

        s(f"Search results for query '{self.query}'")
        if len(self.filter_history) > 0:
            s("With filters: " + ", ".join(self.filter_history))
        s()
        for i, b in enumerate(self.buckets):
            s(f"Bucket {i}: {b.name}")
            sources = set(s.source for s in b.symbols)

            if len(b.symbols) == 0:
                s("No hits", indent=True)
            else:
                s(
                    f"{len(b.symbols)} hits, {len(sources)} sources",
                    indent=True,
                )
                if len(b.symbols) in range(1, 6):
                    s("Hits: " + ", ".join([s.name for s in b.symbols]), indent=True)
                if len(sources) in range(1, 6):
                    s("Sources: " + ", ".join(sources), indent=True)
            s()
        return out

    def p(self) -> None:
        """Convenience method to print str rep."""
        print(self)



"""`SourceType` is a literal type with all known sources; used for type hinting and to
support autocomplete suggestions.
"""

from typing import Literal

SourceType = Literal[
    "yahoo",
    "coingecko",
]



"""Everything related to searching via coingecko."""

import functools
from pycoingecko import CoinGeckoAPI
from ..symbol import Symbol
from . import SearchResult


@functools.lru_cache(maxsize=None)
def get_symbol_map() -> list:
    """Get the symbol map. Separate function to use caching, so the API doesn't get
    hit too often.
    """
    return CoinGeckoAPI().get_coins_list()


def _matches_to_symbols(matches: list) -> list:
    """Convert a list of Coingecko matches to a list of `Symbol`s."""
    return [
        Symbol(
            name=x["symbol"],
            source="coingecko",
            query=x["id"],
            aliases=[x["id"], x["name"]],
        )
        for x in matches
    ]


def coingecko_search(query: str) -> SearchResult:
    """Find coingecko ids that match `query` somehow. Returns `SearchResult`."""
    # Note: No rate limiting here because search is simply a lookup in the symbol map,
    # which gets cached in the first retrieval.
    matches = [
        entry
        for entry in get_symbol_map()
        if query.lower() in entry["name"].lower()
        or query.lower() in entry["symbol"].lower()
    ]
    return SearchResult(query=query, symbols=_matches_to_symbols(matches))



"""Price-related types."""

from typing import NamedTuple
import pandas as pd
from .. import SourceType

PriceHistory = NamedTuple("PriceHistory", [("df", pd.DataFrame), ("currency", str)])
PricePoint = NamedTuple(
    "PricePoint", [("when", pd.Timestamp), ("price", float), ("currency", str)]
)


class SymbolNotFoundError(Exception):
    def __init__(self, source: SourceType, query: str, *args, **kwargs):
        msg = f"No symbol found on source '{source}' for query '{query}'."
        super().__init__(msg, *args, **kwargs)


class CurrencyPreferenceNotFoundError(Exception):
    def __init__(self, source: SourceType, cur_pref: str, *args, **kwargs):
        msg = f"No currency preference found on source '{source}' for '{cur_pref}'."
        super().__init__(msg, *args, **kwargs)



"""Symbol class."""

from typing import Union, Optional, ClassVar
from dataclasses import dataclass, field
import datetime
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from ..price import price_history, price_latest, price_point, PricePoint, PriceHistory
from .. import SourceType

pd.plotting.register_matplotlib_converters()


@dataclass
class Symbol:
    """Symbol class. Encapsulates all the relevant information around a financial symbol
    and some functionality to get price information, display graphs, etc.

    Notes:
    - Price-related functions rely on caching happening on lower levels to be efficient;
      this is fulfilled thanks to the way tessa's caching works.
    - The initializers don't hit the network -- it will only be hit when accessing the
      price methods or related methods such as `currency`.
    """

    name: str
    """Symbol's name and default query."""

    query: Optional[str] = None
    """Use this to use a query that is different than the name. `name` will be used as
    the `query` if it is `None`.
    """

    source: SourceType = "yahoo"
    """The source to query for this symbol. Defaults to "yahoo"."""

    aliases: list[str] = field(default_factory=list)
    """Optional other names this symbol should be found under. (Will not be used as
    additional queries but "internally" in the `matches` method.)"""

    # Class variables:
    currency_preference: ClassVar[str] = "USD"
    """Use this to set the preferred currency to get price information in.

    This is no guarantee and you should always double-check the currency that gets
    returned by any of the `price_*` methods in actuality. This is also the reason why
    this feature is slightly hidden behind an underscore.

    Since this is a class variable, you can set your preference once for all objects.
    """

    def __post_init__(self) -> None:
        """Re/set some attributes."""
        if self.query is None:
            self.query = self.name

    def __str__(self) -> str:
        return f"Symbol {self.name} with source {self.source}"

    def p(self) -> None:
        """Convenience method to print the symbol."""
        print(str(self))

    def to_yaml(self) -> str:
        """Return a YAML representation of this symbol."""
        return f"""
{self.name}:
    query: {self.query}
    aliases: [{", ".join(self.aliases)}]
    source: {self.source}
"""

    def currency(self) -> str:
        """Return currency for this symbol."""
        currency = self.price_latest().currency
        return currency and currency.upper()

    def _create_price_args(self) -> dict:
        """Create a dictionary of arguments that work with tessa's price functions."""
        args = {
            "query": self.query,
            "source": self.source,
            "currency_preference": self.currency_preference,
        }
        return args

    def price_latest(self) -> PricePoint:
        """Return the latest close price."""
        return price_latest(**self._create_price_args())

    def price_history(self) -> PriceHistory:
        """Return a tuple of the full price history."""
        return price_history(**self._create_price_args())

    def price_point(self, when: Union[str, pd.Timestamp]) -> PricePoint:
        """Look up price at given date `when`. Look for the closest point in time if the
        exact point in time is not found.
        """
        return price_point(**self._create_price_args(), when=when)

    def price_graph(self, monthsback: int = 6) -> tuple:
        """Display this symbol's price graph over the last monthsback months.

        Returns from_date, fig, and ax in order for subclass functions to add to the
        information and even the graph displayed here.
        """
        fig, ax = plt.subplots(figsize=(16, 8))
        hist = self.price_history().df

        # Calc start date:
        from_date = datetime.date.today() - pd.offsets.DateOffset(months=monthsback)
        from_date = from_date.strftime("%Y-%m-%d")

        # Plot the prices:
        sns.lineplot(ax=ax, data=hist.loc[from_date:, "close"], marker="o").set(
            title=self.name
        )

        return from_date, fig, ax

    def matches(self, what: str) -> bool:
        """Check if `what` matches this symbol's name or aliases. Also tries to match
        things like SPICHA if SPICHA.SW is in the aliases. Ignores case.
        """
        candidates = [c.lower() for c in [self.name] + self.aliases]
        candidates += [c.split(".")[0] for c in candidates]
        return what.lower() in candidates



"""Source class and related functions."""

from __future__ import annotations
from dataclasses import dataclass
from typing import Callable, Generator, TYPE_CHECKING
import warnings
import requests
from .sourcetype import SourceType
from .rate_limiter import RateLimiter

if TYPE_CHECKING:
    from ..price import PriceHistory


@dataclass
class Source:
    """A `Source` encapsulates everything related to a source such as Yahoo or
    Coingecko.
    """

    get_price_history: Callable
    """Callback to retrieve price history."""

    get_search_results: Callable
    """Callback to retrieve search results."""

    rate_limiter: RateLimiter
    """Rate limiter object."""

    def get_price_history_bruteforcefully(
        self, query: str, currency_preference: str = "USD"
    ) -> PriceHistory:
        """Variant of `get_price_history` that will ignore some server-side errors
        (which often seem to be intermittent, at least on Coingecko) and just
        retries.
        """
        max_tries = 100
        tries = 0
        while True:
            tries += 1
            try:
                return self.get_price_history(query, currency_preference)
            except requests.HTTPError as exc:
                if exc.response.status_code not in [504, 503, 502]:
                    raise exc
                if tries > max_tries:
                    raise ConnectionError(
                        f"Cannot access source, even after {max_tries} tries. "
                        "(Reraising from latest exception, there might have been "
                        "several.)"
                    ) from exc
                warnings.warn(
                    f"Latest request raised for status code {exc.response.status_code}."
                    f" Retrying ({tries}/{max_tries}).",
                    RuntimeWarning,
                )


def get_source(name: SourceType) -> Source:
    """Get a specific source."""
    # Prevent circular dependencies -- pylint: disable=import-outside-toplevel
    from .sources_directory import SOURCES_DIRECTORY

    try:
        return SOURCES_DIRECTORY[name]
    except KeyError as exc:
        raise ValueError(
            f"Unknown source '{name}'. Supported source are: {SOURCES_DIRECTORY.keys()}"
        ) from exc


def get_all_sources() -> Generator:
    """Get all the known sources."""
    # Prevent circular dependencies -- pylint: disable=import-outside-toplevel
    from .sources_directory import SOURCES_DIRECTORY

    for source in SOURCES_DIRECTORY.values():
        yield source


def reset_rate_limiters() -> None:
    """Reset the rate limiters of all sources."""
    for source in get_all_sources():
        source.rate_limiter.reset()



"""Retrieve price information."""

from functools import lru_cache, wraps
from typing import Union
import pandas as pd
from . import PriceHistory, PricePoint
from .. import SourceType
from .. import sources


def custom_cache_wrapper(func):
    """To preserve the function's signature _and_ give access to `cache_clear` etc."""
    cached_func = lru_cache(maxsize=None)(func)
    wrapped_func = wraps(func)(cached_func)
    return wrapped_func


@custom_cache_wrapper
def price_history(
    query: str,
    source: SourceType = "yahoo",
    currency_preference: str = "USD",
) -> PriceHistory:
    """Get price history and return `PriceHistory`, i.e., a tuple of a dataframe with
    the price history and the effective currency. Note that the effective currency
    returned might differ from the currency_preference.

    - `query`: A query string that makes sense in combination with the source. E.g.,
      "BTC-USD" for "yahoo" or "bitcoin" for "coingecko".
    - `source`: The source to query. Defaults to "yahoo".
    - `currency_preference`: The currency to the prices should be returned in; defaults
      to "USD". The effective currency might differ and will be returned in the second
      return value.
    """
    src = sources.get_source(source)
    src.rate_limiter.rate_limit()
    df, effective_currency = src.get_price_history_bruteforcefully(
        query, currency_preference
    )
    return PriceHistory(df.copy(), effective_currency.upper())
    # (Returning a copy of the dataframe so the cached original is preserved even if it
    # gets modified by the caller.)


def price_point(
    query: str,
    when: Union[str, pd.Timestamp],
    source: SourceType = "yahoo",
    currency_preference: str = "USD",
) -> PricePoint:
    """Return the price at a given point in time given by `when`. Look for the closest
    point in time if the exact point in time is not found. Returns a `PricePoint`, i.e.,
    a tuple of the price, the effective timestamp of the price, and the currency.

    Arguments other than `when` are the same as with `price_history`.

    Example call:
    ```
    price_point("AAPL", "2020-01-01")
    ```
    """
    df, currency = price_history(query, source, currency_preference)
    price = df.iloc[df.index.get_indexer([when], method="nearest")[0]]
    return PricePoint(when=price.name, price=float(price), currency=currency)


def price_point_strict(
    query: str,
    when: str,
    source: SourceType = "yahoo",
    currency_preference: str = "USD",
) -> PricePoint:
    """Same as `price_point` but will return either the price at the exact point in time
    or raise a KeyError.
    """
    df, currency = price_history(query, source, currency_preference)
    return PricePoint(
        when=df.loc[when].name, price=float(df.loc[when]["close"]), currency=currency
    )


def price_latest(
    query: str,
    source: SourceType = "yahoo",
    currency_preference: str = "USD",
) -> PricePoint:
    """Same as `price_point` but will return the latest price."""
    df, currency = price_history(query, source, currency_preference)
    return PricePoint(
        when=df.iloc[-1].name, price=float(df.iloc[-1]["close"]), currency=currency
    )



"""The directory (i.e., dictionary) of all know sources. To be accessed via the
functions in `tessa.sources.sources`. This module is factored out so the directory can
be imported late within functions in order to prevent circular dependencies.
"""

from typing import Dict, Final
from .sources import Source
from .sourcetype import SourceType
from .rate_limiter import RateLimiter
from ..price import yahoo as yahooprice, coingecko as coingeckoprice
from ..search import yahoo as yahoosearch, coingecko as coingeckosearch

SOURCES_DIRECTORY: Final[Dict[SourceType, Source]] = {
    "yahoo": Source(
        get_price_history=yahooprice.get_price_history,
        get_search_results=yahoosearch.yahoo_search,
        rate_limiter=RateLimiter(0.5),
    ),
    "coingecko": Source(
        get_price_history=coingeckoprice.get_price_history,
        get_search_results=coingeckosearch.coingecko_search,
        rate_limiter=RateLimiter(2.5),
    ),
}



"""Get Yahoo Finance search results via scraping."""

from typing import List
import pandas as pd
from bs4 import BeautifulSoup
import requests
from .. import Symbol
from ..search import SearchResult
from .. import sources

NUM_PER_PAGE = 1000
URL_BLUEPRINT = "https://finance.yahoo.com/lookup/all?s={}&t=A&b={}&c=" + str(
    NUM_PER_PAGE
)


def create_headers(url: str) -> dict:
    return {
        "authority": "finance.yahoo.com",
        "method": "GET",
        "path": url,
        "scheme": "https",
        "accept": "text/html,application/xml;q=0.9",
        "accept-encoding": "gzip, deflate, br",
        "accept-language": "en-US,en;q=0.9",
        "referer": "https://finance.yahoo.com/",
        "sec-fetch-mode": "navigate",
        "sec-fetch-site": "same-origin",
        "sec-fetch-user": "?1",
        "upgrade-insecure-requests": "1",
        "user-agent": "Mozilla/5.0 (Windows NT 10.0;)",
    }


def get_tables(query: str, offset: int) -> List[pd.DataFrame]:
    """Retrieve search page for query and at offset, parse all the tables into
    dataframes and return that list.
    """
    sources.get_source("yahoo").rate_limiter.rate_limit()
    url = URL_BLUEPRINT.format(query, str(offset))
    page = requests.get(url, headers=create_headers(url))
    soup = BeautifulSoup(page.content, "lxml")
    return [pd.read_html(str(table))[0] for table in soup.find_all("table")]


def get_search_results(query: str) -> pd.DataFrame:
    """Get all search results as a dataframe. Returns an empty dataframe if there are no
    results.
    """
    offset = 0
    df = pd.DataFrame()
    tables = get_tables(query, offset)
    while len(tables) == 1:
        df = pd.concat([df, tables[0]])
        offset += NUM_PER_PAGE
        tables = get_tables(query, offset)

    if len(tables) > 1:
        raise RuntimeError(
            "Yahoo returned more than one table on the search "
            "results page, which is unexpected. Please investigate and/or "
            "submit an issue at github/tessa."
        )

    return df.reset_index(drop=True)


def dataframe_to_symbols(df: pd.DataFrame) -> List[Symbol]:
    """Convert a search result dataframe to a list of `Symbol`s."""
    symbols = []
    for _, row in df.iterrows():
        d = row.to_dict()
        args = {
            "name": d["Symbol"],
            "source": "yahoo",
            "aliases": [d["Name"]],
        }
        symbol = Symbol(**args)
        # FIXME Not the optimal solution bc adding attributes that are unknown to the
        # dataclass Symbol. Add type and exchange as optional attributes to Symbol? Or
        # add a mixin like YahooSymbol or TypedSymbol or so?? Or could/should the
        # Symbols in general be much more lenient around which attributes they accept in
        # addition to the mandatory ones?
        symbol.type = d["Type"]
        symbol.exchange = d["Exchange"]
        symbols.append(symbol)
    return symbols


def yahoo_search(query: str) -> SearchResult:
    """Search on Yahoo Finance."""
    return SearchResult(
        query=query, symbols=dataframe_to_symbols(get_search_results(query))
    )



"""Everything coingecko-related (other than search)."""

import pandas as pd
from pycoingecko import CoinGeckoAPI
from .types import PriceHistory, SymbolNotFoundError, CurrencyPreferenceNotFoundError


def dataframify_price_list(prices: list) -> pd.DataFrame:
    """Turn price list returned by Coingecko API into a pricing dataframe in the form
    that we use.
    """
    df = pd.DataFrame(prices)
    df.columns = ["date", "close"]
    df["date"] = pd.to_datetime(df["date"] / 1000, unit="s", utc=True)
    df = df.set_index("date")
    return df


def get_price_history(query: str, currency_preference: str = "USD") -> PriceHistory:
    """Get price history for a given cryptocurrency."""
    try:
        res = CoinGeckoAPI().get_coin_market_chart_by_id(
            id=query,
            vs_currency=currency_preference,
            days="max",  # FIXME Add some lower default bound similar to yahoo?
            interval="daily",
        )["prices"]
    except ValueError as exc:
        if "invalid vs_currency" in str(exc):
            raise CurrencyPreferenceNotFoundError(
                source="coingecko", cur_pref=currency_preference
            ) from exc
        raise SymbolNotFoundError(source="coingecko", query=query) from exc

    return PriceHistory(dataframify_price_list(res), currency_preference)



"""Everything Yahoo-Finance-related (other than search)."""

import io
import contextlib
import pandas as pd
import yfinance as yf
from .types import PriceHistory, SymbolNotFoundError

START_FROM = "2000-01-01"
"""Adjust this date if you need to get historical data further in the past. Note that
extending this date will lead to increased load on the Yahoo Finance servers.
"""


def get_price_history(
    query: str, currency_preference: str = "USD"  # pylint: disable=unused-argument
) -> PriceHistory:
    """Get price history for a given query. Note that `currency_preference` will be
    ignored since Yahoo Finance returns each ticker in the one currency that is set for
    that ticker.

    Note that yfinance has some strange error behavior, e.g., when a ticker doesn't
    exist: `history()` will return an empty dataframe (but with all the column headers
    as usual) and print "No data found, symbol may be delisted" to stdout.
    """
    stdout = io.StringIO()
    ticker = yf.Ticker(query)
    with contextlib.redirect_stdout(stdout):
        df = ticker.history(start=START_FROM, debug=True)
    if "No data found" in stdout.getvalue():
        raise SymbolNotFoundError(source="yahoo", query=query)

    # Simplify dataframe:
    df = df.copy()
    df = df[["Close"]]
    df.index = pd.to_datetime(df.index, utc=True)
    df.index.name = "date"
    df.rename(columns={"Close": "close"}, inplace=True)

    return PriceHistory(df, ticker.get_history_metadata()["currency"])



"""`CountryName` type."""

from typing import Literal

CountryName = Literal[
    "andorra",
    "united arab emirates",
    "afghanistan",
    "antigua and barbuda",
    "anguilla",
    "albania",
    "armenia",
    "angola",
    "antarctica",
    "argentina",
    "american samoa",
    "austria",
    "australia",
    "aruba",
    "åland",
    "azerbaijan",
    "bosnia and herzegovina",
    "barbados",
    "bangladesh",
    "belgium",
    "burkina faso",
    "bulgaria",
    "bahrain",
    "burundi",
    "benin",
    "saint barthélemy",
    "bermuda",
    "brunei",
    "bolivia",
    "bonaire",
    "brazil",
    "bahamas",
    "bhutan",
    "bouvet island",
    "botswana",
    "belarus",
    "belize",
    "canada",
    "cocos [keeling] islands",
    "democratic republic of the congo",
    "central african republic",
    "republic of the congo",
    "switzerland",
    "ivory coast",
    "cook islands",
    "chile",
    "cameroon",
    "china",
    "colombia",
    "costa rica",
    "cuba",
    "cape verde",
    "curacao",
    "christmas island",
    "cyprus",
    "czech republic",
    "germany",
    "djibouti",
    "denmark",
    "dominica",
    "dominican republic",
    "algeria",
    "ecuador",
    "estonia",
    "egypt",
    "western sahara",
    "eritrea",
    "spain",
    "ethiopia",
    "finland",
    "fiji",
    "falkland islands",
    "micronesia",
    "faroe islands",
    "france",
    "gabon",
    "united kingdom",
    "grenada",
    "georgia",
    "french guiana",
    "guernsey",
    "ghana",
    "gibraltar",
    "greenland",
    "gambia",
    "guinea",
    "guadeloupe",
    "equatorial guinea",
    "greece",
    "south georgia and the south sandwich islands",
    "guatemala",
    "guam",
    "guinea-bissau",
    "guyana",
    "hong kong",
    "heard island and mcdonald islands",
    "honduras",
    "croatia",
    "haiti",
    "hungary",
    "indonesia",
    "ireland",
    "israel",
    "isle of man",
    "india",
    "british indian ocean territory",
    "iraq",
    "iran",
    "iceland",
    "italy",
    "jersey",
    "jamaica",
    "jordan",
    "japan",
    "kenya",
    "kyrgyzstan",
    "cambodia",
    "kiribati",
    "comoros",
    "saint kitts and nevis",
    "north korea",
    "south korea",
    "kuwait",
    "cayman islands",
    "kazakhstan",
    "laos",
    "lebanon",
    "saint lucia",
    "liechtenstein",
    "sri lanka",
    "liberia",
    "lesotho",
    "lithuania",
    "luxembourg",
    "latvia",
    "libya",
    "morocco",
    "monaco",
    "moldova",
    "montenegro",
    "saint martin",
    "madagascar",
    "marshall islands",
    "north macedonia",
    "mali",
    "myanmar [burma]",
    "mongolia",
    "macao",
    "northern mariana islands",
    "martinique",
    "mauritania",
    "montserrat",
    "malta",
    "mauritius",
    "maldives",
    "malawi",
    "mexico",
    "malaysia",
    "mozambique",
    "namibia",
    "new caledonia",
    "niger",
    "norfolk island",
    "nigeria",
    "nicaragua",
    "netherlands",
    "norway",
    "nepal",
    "nauru",
    "niue",
    "new zealand",
    "oman",
    "panama",
    "peru",
    "french polynesia",
    "papua new guinea",
    "philippines",
    "pakistan",
    "poland",
    "saint pierre and miquelon",
    "pitcairn islands",
    "puerto rico",
    "palestine",
    "portugal",
    "palau",
    "paraguay",
    "qatar",
    "réunion",
    "romania",
    "serbia",
    "russia",
    "rwanda",
    "saudi arabia",
    "solomon islands",
    "seychelles",
    "sudan",
    "sweden",
    "singapore",
    "saint helena",
    "slovenia",
    "svalbard and jan mayen",
    "slovakia",
    "sierra leone",
    "san marino",
    "senegal",
    "somalia",
    "suriname",
    "south sudan",
    "são tomé and príncipe",
    "el salvador",
    "sint maarten",
    "syria",
    "swaziland",
    "turks and caicos islands",
    "chad",
    "french southern territories",
    "togo",
    "thailand",
    "tajikistan",
    "tokelau",
    "east timor",
    "turkmenistan",
    "tunisia",
    "tonga",
    "turkey",
    "trinidad and tobago",
    "tuvalu",
    "taiwan",
    "tanzania",
    "ukraine",
    "uganda",
    "u.s. minor outlying islands",
    "united states",
    "uruguay",
    "uzbekistan",
    "vatican city",
    "saint vincent and the grenadines",
    "venezuela",
    "british virgin islands",
    "u.s. virgin islands",
    "vietnam",
    "vanuatu",
    "wallis and futuna",
    "samoa",
    "kosovo",
    "yemen",
    "mayotte",
    "south africa",
    "zambia",
    "zimbabwe",
]



"""
# Searching symbols

Main function is `tessa.search.search.search`, which will return a
`tessa.search.search_result.SearchResult`, offering all kinds of useful functionality.

Example use:

```python
from tessa import search

>>> r = search("roche")                 # search
>>> r.p()                               # print result statistics
>>> r.filter(exchange="EBS").symbols    # filter and get list of resulting symbols

>>> r = search("jenny")                 # another search
>>> r.filter(source="coingecko").symbols # filter for source (i.e., yahoo or coingecko)

>>> r = search("carbon")                # yet another search
>>> r.filter(source="yahoo", type="ETF").symbols # filter for source and type
```

"""

from .search_result import SearchResult
from .search import search



"""
# Retrieving price information

Main functions:

- `tessa.price.price.price_history`: Retrieve the full history of an asset as a
  dataframe.
- `tessa.price.price.price_point_strict`: Get an asset's price at a certain point in
  time. Fail if no price found.
- `tessa.price.price.price_point`: Same, but find the nearest price if the given point
  in time has no
- `tessa.price.price.price_latest`: Get an asset's latest price.


Example use:

```python
>>> from tessa import price_history, price_point, price_latest

>>> df, currency = price_history("AAPL")

>>> price_point("SAP.DE", "2015-01-01")         # will return price at 2015-01-02

>>> price_point_strict("SAP.DE", "2015-01-01")  # will raise a KeyError

>>> price_latest("ethereum", source="coingecko", currency_preference="CHF")

>>> price_latest("ETH-EUR", source="yahoo")

>>> price_latest("ethereum", source="yahoo")    # error b/c symbol not found on yahoo

>>> price_latest("ETH-EUR", source="coingecko") # error b/c symbol not found on coingecko

```

"""

from .types import PriceHistory, PricePoint
from .price import (
    price_history,
    price_point,
    price_point_strict,
    price_latest,
)



"""Rate Limiter -- makes sure we don't hit the APIs too often by waiting by default and
waiting exponentially if a 429 error is encountered.

The goal is to never run into errors in the first place, because some sites like
coingecko.com take substantial time until they allow-list a blocked IP address again.
That is also why we can't use a library such das Tenacity here.

Note that this module is no longer used by the package. But I left it in here anyway,
because it might come in handy for people who use the pycoingecko or some other library
in other ways than through this package.

Usage: Import the module and call `setup_guards()`.
"""

# pylint: disable=eval-used,exec-used,unused-import

import time
import requests
import pendulum

guards = {
    # Additional attributes that will be added dynamically are: last_call, wait_seconds
    "yahoo": {
        "func_name": "requests.post",
        "initial_wait_seconds": 0.2,
        "pattern": "yahoo.com",
    },
    "coingecko": {
        "func_name": "requests.Session.get",
        "initial_wait_seconds": 2.5,
        "pattern": "coingecko.com",
    },
}


def create_guard(func: callable, guard: dict) -> callable:
    """Create a guard."""

    def guarded_func(*args, **kwargs):

        # If the call is for the url pattern, check if the last call was long enough in
        # the past and wait, if not:
        if any(
            guard["pattern"] in x
            for x in args + tuple(kwargs.values())
            if isinstance(x, str)
        ):
            diff = (pendulum.now() - guard["last_call"]).total_seconds()
            if diff < guard["wait_seconds"]:
                time.sleep(guard["wait_seconds"] - diff)
            guard["last_call"] = pendulum.now()

        # Call the original function and return, increase wait_seconds exponentially if
        # a 429 error is encountered:
        try:
            res = func(*args, **kwargs)
        except (requests.HTTPError, ConnectionError) as exc:
            if "429" in exc:
                guard["wait_seconds"] *= 2
            raise exc

        guard["wait_seconds"] = guard["initial_wait_seconds"]  # Reset if no error
        return res

    return guarded_func


def setup_guards() -> None:
    """Set up guards."""
    for guard in guards.values():
        guard["last_call"] = pendulum.parse("1900")
        guard["wait_seconds"] = guard["initial_wait_seconds"]
        guard["func_orig"] = eval(guard["func_name"])
        exec(f"{guard['func_name']} = create_guard({guard['func_name']}, guard)")


# def reset_guards() -> None:
#     """Reset to the originals."""
#     for guard in guards.values():
#         exec(f"{guard['func_name']} = {guard['func_orig']}")
#         # Note that his doesn't work bc `func_orig` gets interpolated to a string
#         # representation, which in turn leads to a syntax error.



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['tessa',
 'tessa.price',
 'tessa.search',
 'tessa.sources',
 'tessa.symbol',
 'tessa.symbol.geo']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=6',
 'beautifulsoup4>=4.11.1,<5.0.0',
 'frozendict>=2.3',
 'matplotlib>=3.5',
 'pendulum>=2.1',
 'pycoingecko>=2.2',
 'seaborn>=0.11',
 'yfinance>=0.2.3,<0.3.0']

setup_kwargs = {
    'name': 'tessa',
    'version': '0.6.0',
    'description': 'Find financial assets and get their price history without worrying about different APIs or rate limiting.',
    'long_description': '\n# tessa – simple, hassle-free access to price information of financial assets 📉🤓📈\n\ntessa is a Python library to help you **easily search asset identifiers** (e.g.,\ntickers) and **retrieve price information** for assets from different sources such as\nYahoo or Coingecko. It takes care of the different APIs, caching, rate limiting, and\nother hassles.\n\ntessa provides a **Symbol class that encapsulates nicely the methods relevant for a\nsymbol**. tessa also provides functionality to **manage collections of symbols**, store\nand load them, and extend their functionality.\n\nFinally, tessa makes sure to be nice to the sites being accessed and tries to **prevent\nusers from being blocked by 429 rate limiting errors** by 1) caching results upon\nretrieval and 2) keeping track of request timestamps and waiting appropriate amounts of\ntime if necessary.\n\n[→ Check out the full documentation. 📖](https://ymyke.github.io/tessa/tessa.html)\n\n\n## How to use\n\nHere\'s a longer example that quickly shows all aspects of the library. Refer to\nsubmodules [symbol](tessa/symbol.html), [search](tessa/search.html), and\n[price](tessa/price.html) for more information.\n\n- Imports:\n\n```python\nfrom tessa import Symbol, SymbolCollection, search\n```\n\n- Create a symbol for MSFT and access some functions:\n\n```python\ns1 = Symbol("MSFT")         # will use "yahoo" as the default source\ns1.price_latest()           # get latest price\n```\n\n- Create another symbol from a bloomberg ticker as it is used by Yahoo Finance:\n\n```python\ns2 = Symbol("SREN.SW")\ns2.price_point("2022-06-30")    # get price at specific point in time\n```\n\n- Create a symbol from the coingecko source with an id as it is used by coingecko:\n\n```python\ns3 = Symbol("bitcoin", source="coingecko")\ns3.price_graph()            # show price graph\n```\n\n- Search for a more crypto ticker on coingecko:\n\n```python\nres = search("GAME")        # search and print search result summary\nfiltered = res.filter(source="coingecko")  # filter results\nfiltered.p()                # print summary of filtered results\nfiltered.buckets[0].symbols # review the best bucket in the filtered results\ns4 = filtered.buckets[0].symbols[2]   # our symbol is the 3rd in that list\ns4.price_history()          # get entire history\n```\n\n- Build a collection of several symbols and use the collection to retrieve symbols:\n\n```python\nsc = SymbolCollection([s1, s2, s3, s4])\nsc.add(Symbol("AAPL"))      # add another one\nsc.find_one("SREN").price_graph()\n```\n\n- Store and load a symbol collection:\n\n```python\nsc.save_yaml("my_symbols.yaml")\nsc_new = SymbolCollection()\nsc_new.load_yaml("my_symbols.yaml")\n```\n\n- Use a different currency preference:\n\n```python\nsc.find_one("game").price_latest()  # will return price in USD\nSymbol.currency_preference = "CHF"\nsc.find_one("game").price_latest()  # will return price in CHF\n```\n\nNote that `currency_preference` will only have an effect with sources that support it.\nIt is supported for Coingecko but not for Yahoo. So you should always verify the\neffective currency you receive in the result.\n\n\n## Data sources\n\ntessa builds on [yfinance](https://pypi.org/project/yfinance/) and\n[pycoingecko](https://github.com/man-c/pycoingecko) and offers **a simplified and\nunified interface**. \n\nWhy these two sources? Yahoo Finance (via yfinance) is fast and offers an extensive\ndatabase that also contains many non-US markets. Coingecko (via pycoingecko) offers\ngreat access to crypto prices. While Yahoo Finance also offers crypto information,\npycoingecko has the advantage that you can have the prices quoted in many more currency\npreferences (a function that is also exposed via tessa).\n\nMore sources can be added in the future. Let me know in the\n[issues](https://github.com/ymyke/tessa/issues) of you have a particular request.\n\n\n## Main submodules\n\n- [symbol](tessa/symbol.html): working with symbols and symbol collections.\n- [search](tessa/search.html): searching the different sources.\n- [price](tessa/price.html): accessing price functions directly instead of via the\n  `Symbol` class.\n- [sources](tessa/sources.html): if you\'d like to add additional sources to the library.\n\n\n## How to install\n\n`pip install tessa`\n\n\n## Prerequisites\n\nSee `pyproject.toml`. Major prerequisites are the `yfinance` and `pycoingecko` packages\nto access finance information as well as the `beautifulsoup4` package to do some\nscraping for searching on Yahoo Finance.\n\n\n## Repository\n\nhttps://github.com/ymyke/tessa\n\n\n## Future Work\n\nThis is an initial version. There are a number of ideas on how to extend. Please leave\nyour suggestions and comments in the [Issues\nsection](https://github.com/ymyke/tessa/issues).\n\n\n## On terminology\n\nI\'m using symbol instead of ticker because a ticker is mainly used for stock on stock\nmarkets, whereas tessa is inteded to be used for any kind of financial assets, e.g. also\ncrypto.\n\n\n## Other noteworthy libraries\n\n- [strela](https://github.com/ymyke/strela): A python package for financial alerts.\n- [pypme](https://github.com/ymyke/pypme): A Python package for PME (Public Market\n  Equivalent) calculation.\n\n\n## On investpy as a data source\n\nTessa used to use the [investpy package](https://github.com/alvarobartt/investpy) as the\nmain source of information until mid 2022 until investing.com introduced Cloudflare,\nwhich broke access by investpy. 😖 It is currently unclear if investpy will be available\nagain in the future. [You can follow the developments in issue\n600.](https://github.com/alvarobartt/investpy/issues/600) The old tessa/investpy code is\nstill available in the [add-symbols-based-on-investpy\nbranch](https://github.com/ymyke/tessa/tree/add-symbols-based-on-investpy).\n',
    'author': 'ymyke',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/ymyke/tessa',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7.13,<3.10',
}


setup(**setup_kwargs)



"""Rate Limiter -- makes sure we don't hit the APIs too often by making sure a minimum
amount of time has elapsed between calls to an API.

The goal is to never run into errors in the first place, b/c some sites take substantial
time until it allow-lists a blocked IP address again. That is also why we can't use a
library such das Tenacity here.
"""

from dataclasses import dataclass
import datetime
import time
import pendulum


VERY_LONG_AGO = pendulum.parse("1900")


@dataclass
class RateLimiter:
    """Encapsulates state and stats of a rate limiter object."""

    wait_seconds: float
    """Enforce this amount of seconds between subsequent calls."""

    last_call: datetime.datetime = VERY_LONG_AGO
    """Keeps track of last call's timestamp."""

    count_all_calls: int = 0
    """Number of total calls."""

    count_limited_calls: int = 0
    """Number of calls that triggered some waiting."""

    def reset(self):
        """Reset state and stats."""
        self.last_call = VERY_LONG_AGO
        self.count_all_calls = self.count_limited_calls = 0

    def rate_limit(self):
        """Enforce the minimum wait time as specified in `wait_seconds`."""
        diff = (pendulum.now() - self.last_call).total_seconds()
        if diff < self.wait_seconds:
            time.sleep(self.wait_seconds - diff)
            self.count_limited_calls += 1
        self.last_call = pendulum.now()
        self.count_all_calls += 1



"""SymbolCollection class."""

from __future__ import annotations
from typing import Optional, Type, Union, List
import collections
import yaml
from . import Symbol


class SymbolCollection:
    """A collection of `Symbol`s. A `Symbol`'s name is treated as a key in a collection
    and this class enforces as much in the `add` method.
    """

    symbols: List[Symbol]
    """The `Symbol`s in the collection."""

    def __init__(self, symbols: Optional[List[Symbol]] = None) -> None:
        """Initializer. Use `symbols` to add symbols immediately and/or use `add` method
        later.
        """
        self.symbols = []
        if symbols is not None:
            self.add(symbols)

    def __iter__(self) -> object:
        """Make class iterable directly."""
        return self.symbols.__iter__()

    def __repr__(self) -> str:
        return f"{self.__class__.__name__} with {len(self.symbols)} symbols"

    def add(self, symbols: Union[Symbol, List[Symbol]]) -> SymbolCollection:
        """Add symbols to the collection. Ensure names are unique."""
        if not isinstance(symbols, list):
            symbols = [symbols]
        if not all(isinstance(x, Symbol) for x in symbols):
            raise ValueError("Please only add Symbol types to a SymbolCollection.")
        all_names = [s.name for s in self.symbols + symbols]
        duplicates = [k for k, v in collections.Counter(all_names).items() if v > 1]
        if duplicates:
            raise ValueError(
                f"Duplicate names ({', '.join(duplicates)}). "
                "Please make sure each name is only used once."
            )
        self.symbols.extend(symbols)
        return self

    def find_one(self, what: str) -> Optional[Symbol]:
        """Find exactly one `Symbol` that matches the query. Raises `ValueError` if
        there's more than 1 match. Returns `None` if there's no match.
        """
        matches = self.find(what)
        if len(matches) > 1:
            raise ValueError(f"Found several matching symbols for '{what}'")
        return matches[0] if matches else None

    def find(self, what: str) -> List[Symbol]:
        """Find all `Symbol`s that match the query."""
        return [s for s in self.symbols if s.matches(what)]

    def load_yaml(self, yaml_file: str, which_class: Type[Symbol] = Symbol) -> None:
        """Load symbols from a YAML file.

        - `yaml_file`: Name/path of file to load.
        - `which_class`: The class to be instantiated, can be used to instatiate
          subclasses of `Symbol`.
        """
        with open(yaml_file, "r", encoding="utf-8") as stream:
            ymldict = yaml.safe_load(stream)
        new_symbols = [which_class(k, **(v or {})) for k, v in ymldict.items()]
        self.add(new_symbols)

    def to_yaml(self) -> str:
        """Return a YAML representation of all symbols."""
        return "".join([s.to_yaml() for s in self.symbols])

    def save_yaml(self, yaml_file: str) -> None:
        """Save symbols to a YAML file."""
        with open(yaml_file, "w", encoding="utf-8") as stream:
            stream.write(self.to_yaml())



"""Unified search."""

from . import SearchResult
from .. import sources


def search(query: str, silent: bool = False) -> SearchResult:
    """Unified search function. Returns a `tessa.search.SearchResult`.

    - `query`: The string to search for. (Note that this query attribute has a different
      semantics to the query attribute in the `Symbol` class.)
    - `silent`: No print output if True.
    """
    res = SearchResult(query, [])
    for source in sources.get_all_sources():
        res.add_symbols(source.get_search_results(query).symbols)
    if not silent:
        res.p()
    return res
